#!/usr/bin/python3
# Copyright (c) 2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
#

import argparse
from argparse_formatter import FlexiFormatter

test_cases = (
    ("Default", argparse.HelpFormatter),
    ("Flexi", FlexiFormatter),
)


def argparse_demo(formatter):
    parser = argparse.ArgumentParser(
        epilog="""
            This is a multi-paragraph epilog. It is presenting data that would
            benefit by being visually broken up into pieces.

            It sure would be nice if it was represented that way.

              1. This is a pretty long line, in a bullet list - getting longer
                 and longer and longer
              2. this isn't
            """,
        formatter_class=formatter,
    )

    parser.add_argument(
        "--arg",
        help="""
            This same feature would be useful for arguments that would benefit
            from more explanation.

              * It looks nicer
              * It is easier to read, even if some of the bullets get to be a little long.
        """,  # noqa
    )

    return parser.format_help()


for (name, formatter) in test_cases:
    print("*************************")
    print("Using the {} formatter".format(name))
    print("*************************")
    print()
    print(argparse_demo(formatter))
    print()



# Copyright (c) 2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
#

from argparse import HelpFormatter
import re as _re
import textwrap

class ParagraphHelpFormatter(HelpFormatter):
    """Help message formatter which respects paragraphs and bulleted lists.

    Only the name of this class is considered a public API. All the methods
    provided by the class are considered an implementation detail.
    """

    def _split_lines(self, text, width):
        return self._para_reformat(text, width)

    def _fill_text(self, text, width, indent):
        lines = self._para_reformat(text, width - len(indent))
        linearray = [indent + line if line else "" for line in lines]
        return "\n".join(linearray)

    def _indents(self, line):
        """Return line indent level and "sub_indent" for bullet list text."""

        indent = len(_re.match(r"( *)", line).group(1))
        list_match = _re.match(r"( *)(([*\-+>]+|\w+\)|\w+\.) +)", line)
        if list_match:
            sub_indent = indent + len(list_match.group(2))
        else:
            sub_indent = indent

        return (indent, sub_indent)

    def _split_paragraphs(self, text):
        """Split text in to paragraphs of like-indented lines."""

        import textwrap

        text = textwrap.dedent(text).strip()
        text = _re.sub("\n\n[\n]+", "\n\n", text)

        last_sub_indent = None
        paragraphs = list()
        for line in text.splitlines():
            (indent, sub_indent) = self._indents(line)
            is_text = len(line.strip()) > 0

            if is_text and indent == sub_indent == last_sub_indent:
                paragraphs[-1] += " " + line
            else:
                paragraphs.append(line)

            if is_text:
                last_sub_indent = sub_indent
            else:
                last_sub_indent = None

        return paragraphs

    def _para_reformat(self, text, width):
        """Reformat text, by paragraph."""

        import textwrap

        lines = list()
        for paragraph in self._split_paragraphs(text):

            (indent, sub_indent) = self._indents(paragraph)

            paragraph = self._whitespace_matcher.sub(" ", paragraph).strip()
            new_lines = textwrap.wrap(
                text=paragraph,
                width=width,
                initial_indent=" " * indent,
                subsequent_indent=" " * sub_indent,
            )

            # Blank lines get eaten by textwrap, put it back
            lines.extend(new_lines or [""])

        return lines


class FlexiFormatter(ParagraphHelpFormatter):
    pass

class FlexiHelpFormatter(ParagraphHelpFormatter):
    pass

class ParagraphFormatter(ParagraphHelpFormatter):
    pass



# Copyright (c) 2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
#

import argparse
import pytest
import re
import textwrap
from collections import namedtuple
from pathlib import Path

from argparse_formatter import ParagraphHelpFormatter

NamedString = namedtuple("NamedString", ["name", "string"])
Case = namedtuple("Case", ["formatter_name", "text_name", "result", "ref"])
Formatter = namedtuple("Formatter", ["name", "obj"])

data_path = Path(__file__).resolve().parent / "data"

wordy_text = "this is wordy text. " * 5
liney_text = "this is wordy text. \n" * 5

help_dict = {
    # detect two normal paragraphs
    "wordy": "{0}\n\n{0}".format(wordy_text),
    # ignore single lfs in the text
    "liney": "{0}\n\n{0}".format(liney_text),
    # tolerate blank lines including white space
    "spacey": "{0}\n  \n{0}".format(wordy_text),
}


formatter_dict = {
    "default": argparse.HelpFormatter,
    "paragraph": ParagraphHelpFormatter,
}


@pytest.fixture(params=[NamedString(x, y) for (x, y) in help_dict.items()])
def helpstr(request):
    return request.param


@pytest.fixture(params=[Formatter(x, y) for (x, y) in formatter_dict.items()])
def formatter(request):
    return request.param


@pytest.fixture()
def testcase(helpstr, formatter):
    parser = argparse.ArgumentParser(
        prog="foo", epilog=helpstr.string, formatter_class=formatter.obj
    )
    parser.add_argument("-foo", help=helpstr.string)
    test_text = parser.format_help()

    data_file_path = data_path / (formatter.name + ".ref")
    # The lazy man's way to create reference data
    # data_file_path.write_text(test_text)

    return Case(
        formatter.name, helpstr.name, test_text, data_file_path.read_text()
    )


def test_case(testcase):
    print(testcase.result)
    assert testcase.result == testcase.ref


SimpleCase = namedtuple("SimpleCase", ["width", "input", "out"])


@pytest.mark.parametrize(
    "case",
    [
        SimpleCase(
            10,
            """
                base text
                  1. text to wrap
            """,
            textwrap.dedent(
                """
                    base text
                      1. text
                         to
                         wrap
                """
            ).strip()
        ),
        SimpleCase(
            10,
            """
                base text
                  - text to wrap
            """,
            textwrap.dedent(
                """
                    base text
                      - text
                        to
                        wrap
                """
            ).strip()
        ),
        SimpleCase(
            10,
            """
                base text
                  1. text to wrap
            """,
            textwrap.dedent(
                """
                    base text
                      1. text
                         to
                         wrap
                """
            ).strip()
        )
    ]
)
def test_flexi_para_reformat(case):
    out = "\n".join(
        ParagraphHelpFormatter("foo")._para_reformat(case.input, case.width)
    )
    print(out)
    assert case.out == out


bullet_template = SimpleCase(
    10,
    """
        base text
          - text to wrap
    """,
    textwrap.dedent(
        """
            base text
              - text
                to
                wrap
        """
    ).strip()
)


@pytest.fixture(params=[x for x in "*-+>"])
def bullet_case(request):
    case = SimpleCase(
        10,
        re.sub("-", request.param, bullet_template.input),
        re.sub("-", request.param, bullet_template.out),
    )

    return case


def test_bullet_chars(bullet_case):
    fmt = ParagraphHelpFormatter("foo")
    lines = fmt._para_reformat(bullet_case.input, bullet_case.width)
    out = "\n".join(lines)
    print(out)
    assert bullet_case.out == out
